@@ -8,6 +8,7 @@
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
 # obtain one at https://mozilla.org/MPL/2.0/.
 
+import abc
 import inspect
 import math
 from random import Random
@@ -16,6 +17,7 @@ from typing import Any
 import attr
 
 from hypothesis.control import should_note
+from hypothesis.internal.conjecture.data import ConjectureData
 from hypothesis.internal.reflection import define_function_signature
 from hypothesis.reporting import report
 from hypothesis.strategies._internal.core import lists, permutations, sampled_from
@@ -23,30 +25,34 @@ from hypothesis.strategies._internal.numbers import floats, integers
 from hypothesis.strategies._internal.strategies import SearchStrategy
 
 
-class HypothesisRandom(Random):
+class HypothesisRandom(Random, abc.ABC):
     """A subclass of Random designed to expose the seed it was initially
     provided with."""
 
-    def __init__(self, note_method_calls):
-        self.__note_method_calls = note_method_calls
+    def __init__(self, *, note_method_calls: bool) -> None:
+        self._note_method_calls = note_method_calls
 
     def __deepcopy__(self, table):
         return self.__copy__()
 
-    def __repr__(self) -> str:
-        raise NotImplementedError
-
+    @abc.abstractmethod
     def seed(self, seed):
         raise NotImplementedError
 
+    @abc.abstractmethod
     def getstate(self):
         raise NotImplementedError
 
+    @abc.abstractmethod
     def setstate(self, state):
         raise NotImplementedError
 
+    @abc.abstractmethod
+    def _hypothesis_do_random(self, method, kwargs):
+        raise NotImplementedError
+
     def _hypothesis_log_random(self, method, kwargs, result):
-        if not (self.__note_method_calls and should_note()):
+        if not (self._note_method_calls and should_note()):
             return
 
         args, kwargs = convert_kwargs(method, kwargs)
@@ -55,9 +61,6 @@ class HypothesisRandom(Random):
         )
         report(f"{self!r}.{method}({argstr}) -> {result!r}")
 
-    def _hypothesis_do_random(self, method, kwargs):
-        raise NotImplementedError
-
 
 RANDOM_METHODS = [
     name
@@ -176,7 +179,7 @@ def normalize_zero(f: float) -> float:
 class ArtificialRandom(HypothesisRandom):
     VERSION = 10**6
 
-    def __init__(self, note_method_calls, data):
+    def __init__(self, *, note_method_calls: bool, data: ConjectureData) -> None:
         super().__init__(note_method_calls=note_method_calls)
         self.__data = data
         self.__state = RandomState()
@@ -184,9 +187,9 @@ class ArtificialRandom(HypothesisRandom):
     def __repr__(self) -> str:
         return "HypothesisRandom(generated data)"
 
-    def __copy__(self):
+    def __copy__(self) -> "ArtificialRandom":
         result = ArtificialRandom(
-            note_method_calls=self._HypothesisRandom__note_method_calls,
+            note_method_calls=self._note_method_calls,
             data=self.__data,
         )
         result.setstate(self.getstate())
@@ -394,7 +397,7 @@ def convert_kwargs(name, kwargs):
 
 class TrueRandom(HypothesisRandom):
     def __init__(self, seed, note_method_calls):
-        super().__init__(note_method_calls)
+        super().__init__(note_method_calls=note_method_calls)
         self.__seed = seed
         self.__random = Random(seed)
 
@@ -407,10 +410,10 @@ class TrueRandom(HypothesisRandom):
         args, kwargs = convert_kwargs(method, kwargs)
         return fn(*args, **kwargs)
 
-    def __copy__(self):
+    def __copy__(self) -> "TrueRandom":
         result = TrueRandom(
             seed=self.__seed,
-            note_method_calls=self._HypothesisRandom__note_method_calls,
+            note_method_calls=self._note_method_calls,
         )
         result.setstate(self.getstate())
         return result
@@ -429,12 +432,12 @@ class TrueRandom(HypothesisRandom):
         self.__random.setstate(state)
 
 
-class RandomStrategy(SearchStrategy):
-    def __init__(self, note_method_calls, use_true_random):
+class RandomStrategy(SearchStrategy[HypothesisRandom]):
+    def __init__(self, *, note_method_calls: bool, use_true_random: bool) -> None:
         self.__note_method_calls = note_method_calls
         self.__use_true_random = use_true_random
 
-    def do_draw(self, data):
+    def do_draw(self, data: ConjectureData) -> HypothesisRandom:
         if self.__use_true_random:
             seed = data.draw_integer(0, 2**64 - 1)
             return TrueRandom(seed=seed, note_method_calls=self.__note_method_calls)
